/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "operator/operator.pmh"

namespace fe
{

void GroupOp::initialize(void)
{
	catalog<String>("Group")="group";
	catalog<String>("Group","label")="Group Name";
	catalog<String>("Group","page")="Define";
	catalog<String>("Group","hint")=
			"The name of the default group to be created."
			"  If no attribute is specified, this is the only group created.";

	catalog<String>("Attribute")="";
	catalog<String>("Attribute","label")="From Attribute";
	catalog<String>("Attribute","page")="Define";
	catalog<String>("Attribute","hint")=
			"Use the string in the named attribute as the group"
			" for each element."
			"  If no attribute is specified or the attribute does not exist"
			" or the attribute is empty for an element,"
			" the Group Name is used.";

	catalog<String>("Element")="Point";
	catalog<String>("Element","page")="Define";
	catalog<String>("Element","choice:0")="Point";
	catalog<String>("Element","choice:1")="Primitive";

	catalog<bool>("EnableBounds")=true;
	catalog<String>("EnableBounds","label")="Enable";
	catalog<String>("EnableBounds","page")="Bounds";
	catalog<bool>("EnableBounds","joined")=true;

	catalog<bool>("Inverse")=false;
	catalog<String>("Inverse","enabler")="EnableBounds";
	catalog<String>("Inverse","page")="Bounds";
	catalog<String>("Inverse","hint")=
		"Include everything the the bounds check would otherwise not include.";

	catalog<String>("Shape")="Box";
	catalog<String>("Shape","enabler")="EnableBounds";
	catalog<String>("Shape","page")="Bounds";
	catalog<String>("Shape","choice:0")="Box";
	catalog<String>("Shape","choice:1")="Sphere";
	catalog<String>("Shape","choice:2")="Surface";

	catalog<SpatialVector>("Center");
	catalog<String>("Center","enabler")="EnableBounds";
	catalog<String>("Center","page")="Bounds";
	catalog<String>("Center","hint")=
			"This is the center of the bounding box or sphere."
			"  For a bounding surface, this will apply a translational"
			" displacement if Displace is turned on.";

	catalog<SpatialVector>("Size")=SpatialVector(1.0,1.0,1.0);
	catalog<String>("Size","enabler")="EnableBounds";
	catalog<String>("Size","page")="Bounds";
	catalog<String>("Size","hint")=
			"This is the dimensions of the bounding box,"
			" the scaling of a unit sphere."
			"  For a bounding surface, this will apply scaling"
			" if Scale is turned on.";

	catalog<bool>("DisplaceSurface")=false;
	catalog<String>("DisplaceSurface","enabler")="Displace";
	catalog<String>("DisplaceSurface","enabler")="EnableBounds";
	catalog<String>("DisplaceSurface","page")="Bounds";
	catalog<bool>("DisplaceSurface","joined")=true;
	catalog<String>("DisplaceSurface","hint")=
			"This permits the translation when using"
			" an arbitrary bounding surface.";

	catalog<bool>("ScaleSurface")=false;
	catalog<String>("ScaleSurface","enabler")="Displace";
	catalog<String>("ScaleSurface","enabler")="EnableBounds";
	catalog<String>("ScaleSurface","page")="Bounds";
	catalog<String>("ScaleSurface","hint")=
			"This permits the scaling when using"
			" an arbitrary bounding surface.";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Bounding Surface");
	catalog<bool>("Bounding Surface","optional")=true;
}

void GroupOp::handle(Record& a_rSignal)
{
	const String attribute=catalog<String>("Attribute");
	const BWORD inverse=catalog<bool>("Inverse");
	const SpatialVector center=catalog<SpatialVector>("Center");
	const SpatialVector size=catalog<SpatialVector>("Size");
	const BWORD displaceSurface=catalog<bool>("DisplaceSurface");
	const BWORD scaleSurface=catalog<bool>("ScaleSurface");

	String shape=catalog<String>("Shape");

	if(catalog<bool>("EnableBounds"))
	{
		if(size[0]<=0.0 || size[1]<=0.0 || size[2]<=0.0)
		{
			if(!inverse)
			{
				return;
			}

			//* inverse nothing = everything
			shape="All";
		}
	}
	else
	{
		if(inverse)
		{
			return;
		}

		//* boundless = everything
		shape="All";
	}

	m_doPrimitive=(catalog<String>("Element")=="Primitive");

	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spElementPosition;
	if(!access(spElementPosition,m_spOutputAccessible,
			m_doPrimitive? e_primitive: e_point,e_position)) return;

	m_spElementAttribute=NULL;
	if(!attribute.empty())
	{
		access(m_spElementAttribute,"Input Surface",
				m_doPrimitive? e_primitive: e_point,attribute,e_warning);
	}

	m_defaultGroup=catalog<String>("Group");
	m_groupMap.clear();

	const U32 elementCount=spElementPosition->count();
	if(shape=="Box")
	{
		const SpatialBox box(center-0.5*size,size);
		for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
		{
			const SpatialVector point=
					spElementPosition->spatialVector(elementIndex);

			if((intersecting(box,point)) != inverse)
			{
				groupElement(elementIndex);
			}
		}
	}
	else if(shape=="Sphere")
	{
		for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
		{
			const SpatialVector point=
					spElementPosition->spatialVector(elementIndex);

			const SpatialVector transformed=2.0*SpatialVector(
					(point[0]-center[0])/size[0],
					(point[1]-center[1])/size[1],
					(point[2]-center[2])/size[2]);

			if((magnitudeSquared(transformed)<1.0) != inverse)
			{
				groupElement(elementIndex);
			}
		}
	}
	else if(shape=="Surface")
	{
		sp<SurfaceI> spBounds;
		if(!access(spBounds,"Bounding Surface")) return;

		spBounds->setRefinement(0);

		const SpatialVector reCenter=
				displaceSurface? center: SpatialVector(0.0,0.0,0.0);
		const SpatialVector reSize=
				scaleSurface? size: SpatialVector(1.0,1.0,1.0);

		for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
		{
			const SpatialVector point=
					spElementPosition->spatialVector(elementIndex);

			const SpatialVector transformed=SpatialVector(
					(point[0]-reCenter[0])/reSize[0],
					(point[1]-reCenter[1])/reSize[1],
					(point[2]-reCenter[2])/reSize[2]);

			const SurfaceI::Containment containment=
					spBounds->containment(transformed);

			if((containment==SurfaceI::e_inside) != inverse)
			{
				groupElement(elementIndex);
			}
		}
	}
	else
	{
		//* boundless

		for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
		{
			groupElement(elementIndex);
		}
	}
}

void GroupOp::groupElement(I32 a_index)
{
	String groupName;
	if(m_spElementAttribute.isValid())
	{
		groupName=m_spElementAttribute->string(a_index);
		if(groupName.empty())
		{
			groupName=m_defaultGroup;
		}
	}
	else
	{
		groupName=m_defaultGroup;
	}

	if(groupName.empty())
	{
		return;
	}

	sp<SurfaceAccessorI>& rspElementGroup=m_groupMap[groupName];
	if(rspElementGroup.isNull() &&
				!access(rspElementGroup,m_spOutputAccessible,
				m_doPrimitive? e_primitiveGroup: e_pointGroup,groupName))
	{
		return;
	}

	rspElementGroup->append(a_index);
}

} // namespace
